// Clock & timer related scripts

// ------------------------- * Clock thread * -------------------------

/*
** clockUpdate
*/
function clockUpdate()
{
	var atime, d, hh, mm, ss, ampm;
	var year, month, date;
	var ms;

// set thread title
	host.setTitle("Clock update");

	while(1)
	{
		d=new Date();
		hh=d.getHours(); mm=d.getMinutes(); ss=d.getSeconds(); // guaranteed integer values

	/* update time on both digital and analog clock */
		if(cfg_clock24hour)
			gui_time.time.setTime(hh, mm, ss, "");
		else if(hh<12)
			gui_time.time.setTime((hh+11)%12+1, mm, ss, "AM");
		else
			gui_time.time.setTime((hh- 1)%12+1, mm, ss, "PM");

	// Date
		year =d.getFullYear();
		month=d.getMonth();
		date =d.getDate();
		gui_time.date.setDate(date, month, year, d.getDay());
	// Analog clock
		ms=d.getTime()%1000;
		if(cfg_clocksmooth)
		{
			gui_time.clock.setTime(hh%12, mm, ss, ms/10);
			host.sleep(25);
		}
		else
		{
			gui_time.clock.setTime(hh%12, mm, ss, 25);
		/* bounce back */
			host.setTimeout(50, function(){gui_time.clock.setTime(this.hh, this.mm, this.ss);}, {hh: hh%12, mm: mm, ss: ss});
			host.sleep(1000-ms);
		}
	}
}

// ------------------------- * Timer object * -------------------------
function Timer(p_countUp, clock)
{
	// private variables
	var countUp = p_countUp; // up or down timer?
	var timerUpdate = false; // is the timer running?
	var timerFreeze = false; // is the display frozen?
	var timerMs = 0;
	var constOffset = 0; // this helps us do a little less arithmetic every cycle
	var oldmins = -1;
	var oldsecs = -1;
	var clock_ring = 0;  // the "last-minute ring" value, should stay at 0 for up-timers
	var fromValue=0; // From value: for up timer it's 0, for down timer it's "Count Down From" XXX: make it member variable?
	const ring_show_secs = 60;  // how many seconds from expiration the ring should be shown static
	const ring_blink_secs = 15; // how many seconds from expiration the ring should be shown blinking
	const ring_clear_ms = 400;  // how many ms in each sec the ring should be cleared to effect blinking

	// public functions

	// timer thread; note that other threads get to change this object's variables only while it's asleep
	this.updateThread = function()
	{
		var sleepTime = 100; // sleepTime is adjusted below to try to make TimerMs hit exact tenths of a sec

		// set thread title
		host.setTitle("Timer update");

		while (1)
		{
			if (!timerUpdate)
			{
				host.sleep(100);
				continue;
			}

			if (countUp)
			{
				timerMs = constOffset + host.getms();
				sleepTime = 100 - timerMs % 100;

				if (timerMs > timerMaxValue) // reset if more than 3 digits
				{
					timerUpdate = false;
					this.reset();
				}
			}
			else
			{
				timerMs = constOffset - host.getms();
				sleepTime = 1 + timerMs % 100;

				if(timerMs<=0)
				{
				// reset to default and stop
					timerUpdate = false;
					sendGPIOcommand(dtmr_lpid, 7, gpio.pulse); // send STOP pulse
					this.setMs(fromValue);
					sleepTime=100; // otherwise it can remain negative
				}
				else
					updateRing();
			}

			if(timerFreeze===false) this.updateDisp();
			host.sleep(sleepTime);
		}
	};

	this.updateDisp=function()
	{
		var msec=(timerFreeze===false)?timerMs:timerFreeze;
		var mins=Math.floor(msec/60000);
		var secs=Math.floor((msec%60000)/1000);
		var tenths=Math.floor((msec%1000)/100);
		var show_tenths=timerFreeze!==false || !timerUpdate || showTenths(); // display tenths if timer is paused

		if(!countUp && !show_tenths && tenths!=0)
		{
		// if tenths are not shown, round then up: 58.7 => 59, 58.1 => 59, 58.0 => 58
			if(++secs==60)
			{
				mins++;
				secs=0;
			}
		}

		if(show_tenths)
			this.setTime(mins, secs, tenths);
		else if(secs!=oldsecs || mins!=oldmins) // redisplay min/secs only if they changed TODO: check redisplay on tenth hide
		{
			this.setTime(mins, secs, -1);
			oldsecs=secs;
			oldmins=mins;
		}
	};

	this.start=function()
	{
		if(timerUpdate) return;
		computeOffset();
		oldmins=-1;
		oldsecs=-1;
		sendGpio();
		timerUpdate=true;
		if(!countUp) updateRing();
		lwSendTimerEvent(countUp, {Value:timerMs/1000.0, State:timerUpdate?"ON":"OFF", Freeze:(timerFreeze===false)?"OFF":timerFreeze/1000.0});
		sendGPIOcommand(countUp?utmr_lpid:dtmr_lpid, 8, gpio.pulse);
	};
	this.stop=function()
	{
		if(!timerUpdate) return;
		timerUpdate=false;
		lwSendTimerEvent(countUp, {Value:timerMs/1000.0, State:timerUpdate?"ON":"OFF", Freeze:(timerFreeze===false)?"OFF":timerFreeze/1000.0});
		sendGPIOcommand(countUp?utmr_lpid:dtmr_lpid, 7, gpio.pulse);
	// if we caught the ring during off-time of the blinking period
		if(!countUp && timerMs<ring_blink_secs*1000 && clock_ring==0)
			gui_time.clock.setRing(clock_ring=getRingVal());

		this.updateDisp();
	};
	this.toggleStart=function()
	{
		if(timerUpdate)
			this.stop();
		else
			this.start();
	};

	this.freeze=function()
	{
		if(timerFreeze!==false) return;
		timerFreeze=timerMs; // capture current Ms
		lwSendTimerEvent(countUp, {Value:timerMs/1000.0, State:timerUpdate?"ON":"OFF", Freeze:(timerFreeze===false)?"OFF":timerFreeze/1000.0});
		if(countUp)
		{
		// set indication
			mod.setMasterKeyLed(keys.MstTimerPause, 255);
			mod.setMasterKeyLed(keys.CmbTimerPause, 255);
			sendGPIOcommand(utmr_lpid, 5, gpio.on);
		}
		else
			sendGPIOcommand(dtmr_lpid, 5, gpio.on);
	};
	this.setFreezeMs=function(msec)
	{
		if(timerFreeze===msec) return;
		timerFreeze=msec;
		if(!timerUpdate) this.updateDisp(); // show any new manually set values [XXX: is this required?]
		lwSendTimerEvent(countUp, {Value:timerMs/1000.0, State:timerUpdate?"ON":"OFF", Freeze:(timerFreeze===false)?"OFF":timerFreeze/1000.0});
		if(countUp)
		{
		// set indication
			mod.setMasterKeyLed(keys.MstTimerPause, 255);
			mod.setMasterKeyLed(keys.CmbTimerPause, 255);
			sendGPIOcommand(utmr_lpid, 5, gpio.on);
		}
		else
			sendGPIOcommand(dtmr_lpid, 5, gpio.on);
	};
	this.unfreeze=function()
	{
		if(timerFreeze===false) return;
		if(!timerUpdate) this.updateDisp(); // show any new manually set values [XXX: is this required?]
		timerFreeze=false;
		lwSendTimerEvent(countUp, {Value:timerMs/1000.0, State:timerUpdate?"ON":"OFF", Freeze:(timerFreeze===false)?"OFF":timerFreeze/1000.0});
		if(countUp)
		{
		// set indication
			mod.setMasterKeyLed(keys.MstTimerPause, 0);
			mod.setMasterKeyLed(keys.CmbTimerPause, 0);
			sendGPIOcommand(utmr_lpid, 5, gpio.off);
		}
		else
			sendGPIOcommand(dtmr_lpid, 5, gpio.off);
	};

	this.setMs=function(msec)
	{
		timerMs=msec;

		if(timerUpdate) // it's actively running, so really force the new value
			computeOffset();

		if(timerFreeze===false) this.updateDisp();
		lwSendTimerEvent(countUp, {Value:timerMs/1000.0, State:timerUpdate?"ON":"OFF", Freeze:(timerFreeze===false)?"OFF":timerFreeze/1000.0});

		// for stopped downtimers consider showing/hiding the ring, otherwise it ain't worth bothering with
		if(!countUp && !timerUpdate)
		{
			// hide analog count down ring when clock is reset. Show it only when it starts running.
			// in count down case, setMs is used to reset. That is, to set to preset time
			gui_time.clock.setRing(clock_ring=0);
		}

		if(timerUpdate)
			sendGpio();
	};

	this.getValue=function() {return timerMs/1000.0;};
	this.getState=function() {return timerUpdate;};
	this.getFreeze=function() {return (timerFreeze===false)?"OFF":timerFreeze/1000.0;};

	this.reset=function()
	{
		this.setMs(fromValue);
		sendGPIOcommand(countUp?utmr_lpid:dtmr_lpid, 6, gpio.pulse);
	};

	this.updateTMRT=function()
	{
		this.updateDisp();
	};
	this.setFromValue=function(ms)
	{
		if(ms!=fromValue)
		{
			if('setFromLabel' in this) this.setFromLabel(ms);
			lwSendTimerEvent(countUp, {FromVal:ms});
			fromValue=ms;
		}
	};

	// private functions
	function updateRing()
	{
		var r = getRingVal(); // should the ring be shown or cleared?

		// if it looks like it should be shown but it's during the blinking period's off-phase
		if (r && timerMs < ring_blink_secs * 1000 && (timerMs - 100) % 1000 >= ring_clear_ms){
			r = 0; // so don't show it
		}

		if (r != clock_ring) // if it's different from what we last had
			gui_time.clock.setRing(clock_ring = r); // then it's worth redrawing it or clearing, depending on r
	}

	// figures out what the ring value ought to be, but doesn't consider the blinking period
	function getRingVal()
	{
		if (timerMs <= ring_show_secs * 1000)
			// here we can't do simple ceil as we need to look at millisecond resolution
			// that is, 58.045 should be treated as 58, not 59 as it has millisecond component zero (58.0)
			return Math.ceil(Math.floor(timerMs / 100) / 10);
		else
			return 0;
	}

	// fix the relationship between timerMs and host.getms(); this is more important than timerMs itself
	function computeOffset()
	{
		var ms_snapshot = host.getms();
		constOffset = timerMs + (countUp ? -ms_snapshot : ms_snapshot);
	}

	function sendGpio()
	{
		if(stmo_lpid!=0)
		{
		// send GPIO command if "started from beginning"
			if(timerMs==fromValue)
				sendGPIOcommand(stmo_lpid, countUp?gpio.pin_up_timer:gpio.pin_down_timer, gpio.pulse);
		}
	}

	function showTenths()
	{
		//show on both OR show on UP and this is UP OR show on DOWN and this is DOWN timer
		return (timer_tenth_mode==3 || (countUp && timer_tenth_mode==2) || (!countUp && timer_tenth_mode==1));
	}
}

// ------------------------- * GUI * -------------------------

function createDownTimer(f, x, y)
{
	dnTimerObj=new Timer(false);

	if(cfg_fusionUI)
	{
		dnTimerObj.MS=gui.newLabel(f, x+26, y+4, 130, 34, '', style.tmdn);
		dnTimerObj.setTime=function(mm, ss, tenths)
		{
			if(tenths<0)
				this.MS.setText(host.sprintf("%02u:%02u", mm, ss));
			else
				this.MS.setText(host.sprintf("%02u:%02u.%u", mm, ss, tenths));
		};
	// count down from label
		dnTimerObj.lblCountFrom=gui.newLabel(f, 8, y+38, 312-2*8, 16, '', {font:fontLight(12), color:color.lyellow, align:gui.CENTER, valign:gui.MIDDLE});
		dnTimerObj.setFromLabel=function(ms)
		{
			this.lblCountFrom.setText(host.sprintf("COUNTDOWN FROM %u:%02u", ms/60000, (ms%60000)/1000));
		};
	}
	else
	{
		dnTimerObj.MS=gui.newLabel(f, x+17, y, 125, 28, '', style.tmdn);
		dnTimerObj.setTime=function(mm, ss, tenths)
		{
			if(tenths<0)
				this.MS.setText(host.sprintf("%02u : %02u", mm, ss));
			else
				this.MS.setText(host.sprintf("%02u : %02u .%u", mm, ss, tenths));
		};
	// count down from label
		dnTimerObj.lblCountFrom=gui.newLabel(f, x+ 20, y+28, 120, 14, '', style.tmdn_aux);
		dnTimerObj.setFromLabel=function(ms)
		{
			this.lblCountFrom.setText(host.sprintf("Countdown from %u:%02u", ms/60000, (ms%60000)/1000));
		};
	}
	dnTimerObj.setFromValue(cfg_countdownfrom);
	dnTimerObj.setMs(cfg_countdownfrom);

// launch timer thread
	host.setTimeout(100, dnTimerObj.updateThread, dnTimerObj);
}

function createUpTimer(f, x, y)
{
	upTimerObj=new Timer(true);

	if(cfg_fusionUI)
	{
		upTimerObj.MS=gui.newLabel(f, x+26, y+4, 130, 34, '', style.tmdn);
		upTimerObj.setTime=function(mm, ss, tenths)
		{
			if(tenths<0)
				this.MS.setText(host.sprintf("%02u:%02u", mm, ss));
			else
				this.MS.setText(host.sprintf("%02u:%02u.%u", mm, ss, tenths));
		};
	// timer mode indicator
		timer_mode_ind=gui.newLabel(f, x-55, y+38, 112, 16, '', {font:fontLight(12), color:color.lyellow, align:gui.LEFT, valign:gui.MIDDLE});
		timer_mode_ind.setMode=function(mode)
		{
			switch(mode)
			{
			case 0: this.setText("MANUAL"); break;
			case 1: this.setText("AUTO RESET"); break;
			case 2: this.setText("AUTO ADD"); break;
			}
		};
	}
	else
	{
		upTimerObj.MS=gui.newLabel(f, x+17, y, 125, 28, '', style.tmup);
		upTimerObj.setTime=function(mm, ss, tenths)
		{
			if(tenths<0)
				this.MS.setText(host.sprintf("%02u : %02u", mm, ss));
			else
				this.MS.setText(host.sprintf("%02u : %02u .%u", mm, ss, tenths));
		};
	// timer mode indicator
		timer_mode_ind=
		{
			dot: gui.newLabel(f, x+10, y+28,  8, 17, '', style.tmup_dot),
			lbl: gui.newLabel(f, x+20, y+28, 120, 14, '', style.tmup_aux),
			setMode: function(mode)
			{
				this.dot.setText(".");
				switch(mode)
				{
				case 0: this.lbl.setText("Manual"); break;
				case 1: this.lbl.setText("Auto Reset"); break;
				case 2: this.lbl.setText("Auto Add"); break;
				}
			}
		};
	}
	upTimerObj.setMs(0);
	timer_mode_ind.setMode(timer_mode);

	//TODO add info indicator (design example shows Zephyr 02)
// launch timer thread
	host.setTimeout(100, upTimerObj.updateThread, upTimerObj);
}
